#!/usr/bin/perl -w

#
# response.pl version 1.0
# Written by:
# Aaron Turner <aturner@pobox.com> aka <aturner@vicinity.com>
#
# Based on sam_alert.sh v2.0 by:
# Lance Spitzner lance@spitzner.net and Aaron Turner
#
# Parse, log, and react to FW-1 User Defined Alerts
# For more information, please review the README
#
# Distributed under the terms of this General Public License v2.0
# or greater.  Please see the included file COPYING for details, or
# visit: http://www.gnu.org/copyleft/gpl.html	
#
# This code, and future revisions can be found at:
# http://www.pobox.com/~aturner/fw-respose/

use strict; # for good code
use Mail::Mailer; # for sending email
use Socket; # for DNS lookups
use diagnostics; # better debugging of code

#################################################################
# 		          BEGIN CODE			        #
#################################################################

# Unless you really know what you're doing, you don't want to 
# edit anything below this.  If you're looking for options to
# configure, edit response.cfg

# ############# G L O B A L   V A R I A B L E S #################

my ($message, %headers); # message to be emailed to admin

# Declare Basic Options
use vars qw/$dir $fw1 $email $from $limit $notify_remote_admin $email_server/;

# Declare Advanced Options
use vars qw/$sam_service_block $timeout @sam_services @sam_protect_list @friends $wait_for_limit $sam_options $sam_inhibit $sam_service_ignore_limit $enable_sam_dynamic_enforce @sam_dynamic_enforce $sam_dynamic_enforce_default/;

# Declare data content
use vars qw/$alert_send $sam_split $service_sam_block_send/;

# Temp
use vars qw/$src $ipaddr $dst/;

my $debug = 0; # set to 1 to turn on debuging, else 0

# ############# P R O C E S S   C O N F I G #####################

my $config;  # string containing the config file
open (CONFIG, "response.cfg") || die ("Unable to open file 'response.cfg' in current working directory!\n$!\n");
while (<CONFIG>) {
  $config .= $_ if ($_ =~ /^(\$|\@|\%)/); # only read in the varaibles- skip blanks and comments
}
close (CONFIG);
eval "$config" or die ("Unable to process the response.cfg\n$!"); # eval will return 1 for sucess, 0 for failure.

$headers{to} = $email;
$headers{from} = $from;

# ############# P R O C E S S   I N P U T #######################


my $input = <STDIN>; # read line from stdin
my @input = split(/\s+/, $input); # put each var into the array.
my $date = $input[0];
my $time = $input[1];
my $action = $input[2];
my $firewall_log = $input[3];
my $proto = $input[7];
my $source = $input[9];
my $dest = $input[11];
my $nat_check = 0;
my $icmp_check = 0;
my $service; # set this below
my $srcport = $input[12]; 

# check to see if any NAT or if this is ICMP
# no idea if this works for ICMP over NAT.  Prolly not.
$nat_check = 1 if ($input =~ / \(Valid Address\) /);
$icmp_check = 1 if ($input =~ / icmp /);

# determine the service based on above
if (! $nat_check) {
  if (! $icmp_check) {
    $service = $input[13];
  } else {
    for (14..17) {
      $service .= $input[$_];
    }
  }
} else {
  if (! $icmp_check) {
    $service = $input[15];
  } else {
    for (16..19) {
      $service .= $input[$_];
    }
  }
}

my $scan_number = 0;

open(LOG, "${dir}alert.log") or die("Unable to open log file.  $!");
while (<LOG>) {
  $scan_number ++ if (/$source/);
}
close (LOG);
my $scan = ++ $scan_number;

# Check number of scans and if we should continue sam'ing after 
# we reach the limit.  If limit reached, and no sam, then log
# the intrusion, and exit.
if ( ($scan > $limit) && (! &True_False($sam_service_ignore_limit) ) ) {
  open(LOG,">>${dir}alert.log") || &leave("Unable to open ${dir}alert.log.  $!");
  print LOG $input;
  close(LOG);
  exit 10;
}

# Check to see if -f was used in sam_options.  It may be a problem
my $using_f = "false";
$using_f = "true" if ($sam_options =~ /-f/);

my ($firewall, $enforce);
# Determine if we enabled dynamic enforment of firewalls, if so, determine which
# Firewall object to use.  Else select the default ($sam_dynamic_enforce_default).
if ( (&True_False($enable_sam_dynamic_enforce)) && (! &True_False($using_f)) ) {
  $firewall = &Dynamic_Enforce($firewall_log, \@sam_dynamic_enforce);
  $enforce = "-f $firewall";
} else { # use default
  $firewall = $sam_dynamic_enforce_default;
}



# ###################### F U N C T I O N S ######################

# True_False
###########
# returns 1 if 1, true, or yes are passed.  0 if other.
######################################################

sub True_False {
  my @ok = (1, 'true', 'yes'); # case sensitive list of true values

  foreach (@ok) {
    return 1 if ($_ eq shift);
  }
  return 0; # default value
} # True_False

# Dynamic_Enforce
###################
# Takes in a string (Hostname of firewall) and a string of /matches/replaces/ 
# and runs a number of substitutions based on the /m/r/
##############################################################################

sub Dynmamic_Enforce {
  my $host = shift;  
  my $subs = shift; # ref to a array of subs


  # foreach /<firewall-name>/<firewall-obj>/arg do
  foreach (@$subs) {
    my ($null, $regexp, $fw_obj, $args) = split (/\//, $_);
    
    if ($args =~ /i/) {
      return $fw_obj if ( $host =~ /$regexp/i ); # exit sub if match
    } else {
       return $fw_obj if ( $host =~ /$regexp/ ); # exit sub if match
    }
} # foreach
  # fail back to the default
  return ($sam_dynamic_enforce_default);
} # Dynamic_Enforce

# Resolve
#########
# This function takes in a string, determines if it is an IP,
# if so, does a nslookup, and returns the hostname if any.
# If no resolve or not an IP, return 0.
##############################################################################

sub Resolve {
  my $IP = shift;
  my $name; # return val
  if ($IP =~ /([0-9]{1,3}\.){3}[0-9]{1,3}/) { # not a perfect match, but good enough
    $name = gethostbyaddr(inet_aton($IP), AF_INET) or $name = "NULL";
  }
  return $name if ($name ne 'NULL');
  return 0;
} # Resolve

# Reverse_Resolve
#################
# opposite of &Resolve.  Does name -> IP.
#########################################

sub Reverse_Resolve {
  my $name = shift;
  my $IP; # return val
  if ($name = /[a-zA-Z]*/) {
    $IP = inet_ntoa(inet_aton($name)) or $IP = "NULL";
  }
  return $IP ? ($IP ne 'NULL') : return 0;
} # Reverse_Resolve

# Service_Block
###############
# Port of ServBlk
# This function will block scanners according to the advanced options
# ############################################################################

sub Service_Block {
  my $sam_protect_list_match = "false"; # didn't SAM by default

  foreach my $sam_service (@sam_services) {
    if ($sam_service eq $service) {
      foreach my $sam_protect (@sam_protect_list) {
	my $output = `${fw1}fw sam $sam_options $enforce -t $timeout $sam_inhibit srv $src $sam_protect $service $proto 2>&1`;
	$headers{subject} = "#### SERVICE BLOCK ${src}:${service} ####";
	$sam_protect_list_match = "true";
      }
	
    } 
  } # foreach
  return $sam_protect_list_match; # did we sam or not?
} # Service_Block

# Send_Email
############
# takes in a pointer to the headers (hash), and a string for the body and mail server
# last is method to send the email.  Should be either smtp (send) or test (debuging)
#####################################################################################

sub Send_Email {
  my $headers = shift; # pointer to hash
  my $body = shift;
  my $server = shift;
  my $method = shift;

  my $mailer = new Mail::Mailer $method, Server => $server;
  $mailer->open($headers); # pass $headers by reference
  print $mailer $body; # print the body[
  $mailer->close or return 0; # send email
  return 1;
} # Send_Email

# Bin2Dec
#########
# Converts Binary to Decimal
############################

sub Bin2Dec { 
  return unpack("N", pack("B32", substr("0" x 32 . shift, -32)));
} # Bin2Dec

# Dec2Bin
#########
# Converts Decimal to Binary.
# 2nd argument is number of significant bits
# to return (adding padding or choping off 
# as necessary)  Use 0 if you want to return
# the binary with no padding or chopping, 
# which will return a dynamic length based on 
# big/little endian and 32/64bit OS or something
# like that.  Who knows?  I don't.
#############################################

sub Dec2Bin {
  # do the conversion
  my $str = unpack("B32", pack("N", shift));
  
  # how many bits do we return? 
  my $bits = shift;

  # if user set bits to 0, return with no padding or chopping
  if ($bits == 0) {
    $str =~ s/^0+(?=\d)//;
    return $str;
  }

  # Put $str into an array so we can see how long it is later
  my @str = split(//, $str);

  # If size of the $str != $bits then we need to chop or pad it
  if ($bits < scalar(@str)) {
    my $start = $bits;
    $start *= -1;
    $str = substr($str, $start, $bits);
  } elsif ($bits > scalar(@str)) { 
    # Pad with leading zero's if necessary
    unshift(@str, 0) while ($bits > scalar(@str));
    $str = join("", @str); # put things back into a string.
  }

 return $str;
} # Dec2Bin

# Inside_CIDR
#############
# Determines if an IP address is inside a CIDR block
# $IP, $CIDR
# returns 1 for yes, 0 for no
###############################################################################

sub Inside_CIDR {
  # first get the IP, and then the CIDR
  my ($ip_a, $ip_b, $ip_c, $ip_d) = split (/\./, shift);

  # Split the network from the mask
  my ($net, $mask) = split (/\//, shift);

  # not split each octet
  my @net = split (/\./, $net);

  # if they didn't give us all 4 octets, insert zero's in their place
  push (@net, 0) while (scalar(@net) < 4);
  
  my ($cidr_a, $cidr_b, $cidr_c, $cidr_d) = @net; 
  
  # convert the IP's to their binary equivalents and store them as a single 32 bit string
  my $cidr = &Dec2Bin($cidr_a, 8) .  &Dec2Bin($cidr_b, 8) . &Dec2Bin($cidr_c, 8) . &Dec2Bin($cidr_d, 8);
  my $ip = &Dec2Bin($ip_a, 8) . &Dec2Bin($ip_b, 8) . &Dec2Bin($ip_c, 8) . &Dec2Bin($ip_d, 8);
  
  # overlay the mask
  $ip = substr($ip, 0, $mask);
  $cidr = substr($cidr, 0, $mask);

  # are we inside the CIDR, return our answer
  return 1 ? ($ip == $cidr) : return 0;
  
} # Inside_CIDR

